// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "UICanvas.hpp"
#include "display/pass/RenderPass.hpp"
#include "display/shader/UIShader.hpp"

#include "display/mesh/QuadMesh.hpp"

using namespace Dawn;

void UICanvas::onInit() {
  mesh = std::make_shared<Mesh>();
  mesh->createBuffers(
    QUAD_VERTICE_COUNT * UI_SHADER_QUAD_COUNT,
    QUAD_INDICE_COUNT * UI_SHADER_QUAD_COUNT
  );

  for(int32_t i = 0; i < UI_SHADER_QUAD_COUNT; i++) {
    QuadMesh::bufferWithIndex(
      mesh,
      glm::vec4(0, 0, 1, 1),
      glm::vec4(0, 0, 1, 1),
      i * QUAD_VERTICE_COUNT,
      i * QUAD_INDICE_COUNT,
      i * QUAD_VERTICE_COUNT
    );
  }
}

void UICanvas::onDispose() {
  mesh = nullptr;
}

std::vector<std::shared_ptr<IRenderPass>> UICanvas::getPasses(
  struct RenderPassContext &ctx
) {
  std::vector<std::shared_ptr<IRenderPass>> passes;
  std::unordered_map<shadertexturebinding_t, std::shared_ptr<Texture>> textures;
  UIShaderData data = {
    .projection = ctx.camera->getProjection(),
    .view = ctx.camera->getItem()->getWorldTransform(),
    .model = this->getItem()->getWorldTransform()
  };
  
  size_t quadCount = 0;
  auto itComponents = components.begin();
  while(itComponents != components.end()) {
    auto component = *itComponents;

    // Get this components' quads.
    auto quads = component->getQuads(glm::mat4(1.0));
    for(auto quad : quads) {
      data.quads[quadCount++] = quad;
      assertTrue(quadCount <= UI_SHADER_QUAD_COUNT, "Too many UI quads!");
    }
    ++itComponents;
  }

  // No render passes if no quads.
  if(quadCount == 0) return passes;

  auto pass = createRenderPass<UIShader, UIShaderData>(
    std::ref(*this),
    data,
    textures,
    mesh,
    MeshDrawMode::TRIANGLES,
    0, quadCount * QUAD_INDICE_COUNT
  );
  passes.push_back(pass);


  return passes;
}