// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "display/shader/UIShader.hpp"
#include "util/Macro.hpp"

using namespace Dawn;

void UIShader::getStages(
  const enum ShaderOpenGLVariant variant,
  const struct UIShaderData *rel,
  std::vector<std::shared_ptr<ShaderStage>> &stages,
  std::vector<struct ShaderParameter> &parameters,
  std::vector<struct IShaderStructure> &structures
) {
  // Stages
  std::shared_ptr<ShaderStage> vertex;
  std::shared_ptr<ShaderStage> fragment;

  switch(variant) {
    case ShaderOpenGLVariant::GLSL_330_CORE:
      vertex = std::make_shared<ShaderStage>(
        ShaderStageType::VERTEX,
        "#version 330 core\n"
        "layout (location = 0) in vec3 aPos;\n"
        "layout (location = 1) in vec2 aTexCoord;\n"
        "uniform mat4 u_Projection;\n"
        "uniform mat4 u_View;\n"
        "uniform mat4 u_Model;\n"
        "struct UIShaderQuad {\n"
          "mat4 transform;\n"
          "vec4 quad;\n"
          "vec4 uv;\n"
          "vec4 color;\n"
        "};\n"
        "layout (std140) uniform ub_Quad {\n"
          "UIShaderQuad quads[" MACRO_STRINGIFY(UI_SHADER_QUAD_COUNT) "];\n"
        "};\n"
        "out vec2 o_TextCoord;\n"
        "out vec4 v_Color;\n"
        "void main() {\n"
          "vec4 pos;\n"
          "vec2 coord;\n"
          "int index = int(aPos.z);\n"
          "int quadIndex = index / 4;\n"
          "int vertexIndex = index % 4;\n"
          "struct UIShaderQuad quad = quads[quadIndex];\n"
          "if(vertexIndex == 0) {\n"
            "pos.x = quad.quad.x;\n"
            "pos.y = quad.quad.y;\n"
            "coord.x = quad.uv.x;\n"
            "coord.y = quad.uv.y;\n"
          "} else if(vertexIndex == 1) {\n"
            "pos.x = quad.quad.z;\n"
            "pos.y = quad.quad.y;\n"
            "coord.x = quad.uv.z;\n"
            "coord.y = quad.uv.y;\n"
          "} else if(vertexIndex == 2) {\n"
            "pos.y = quad.quad.w;\n"
            "pos.x = quad.quad.x;\n"
            "coord.x = quad.uv.x;\n"
            "coord.y = quad.uv.w;\n"
          "} else if(vertexIndex == 3) {\n"
            "pos.x = quad.quad.z;\n"
            "pos.y = quad.quad.w;\n"
            "coord.x = quad.uv.z;\n"
            "coord.y = quad.uv.w;\n"
          "}\n"
          "pos.z = 0;\n"
          "pos.w = 1;\n"
          "gl_Position = u_Projection * u_View * u_Model * quad.transform * pos;\n"
          "o_TextCoord = coord;\n"
          "v_Color = quad.color;\n"
        "}"
      );
      
      fragment = std::make_shared<ShaderStage>(
        ShaderStageType::FRAGMENT,
        "#version 330 core\n"
        "in vec2 o_TextCoord;\n"
        "in vec4 v_Color;\n"
        "out vec4 o_Color;\n"
        "void main() {\n"
          "o_Color = v_Color;\n"
        "}\n"
      );
      break;

    default:
      assertUnreachable("Unsupported ShaderOpenGLVariant");
  }

  // Add stages
  stages.push_back(vertex);
  stages.push_back(fragment);

  // Parameters
  parameters.push_back(ShaderParameter(
    "u_Projection",
    &rel->projection,
    ShaderParameterType::MAT4
  ));

  parameters.push_back(ShaderParameter(
    "u_View",
    &rel->view,
    ShaderParameterType::MAT4
  ));

  parameters.push_back(ShaderParameter(
    "u_Model",
    &rel->model,
    ShaderParameterType::MAT4
  ));
  
  structures.push_back(ShaderStructure<struct UIShaderQuad>(
    "ub_Quad",
    &rel->quads,
    ShaderOpenGLStructureType::STD140,
    [&](const struct UIShaderQuad &rel, std::vector<struct ShaderParameter> &parameters) {
      parameters.push_back(ShaderParameter(
        "u_Transform",
        &rel.transform,
        ShaderParameterType::MAT4
      ));

      parameters.push_back(ShaderParameter(
        "u_Quad",
        &rel.quad,
        ShaderParameterType::VEC4
      ));

      parameters.push_back(ShaderParameter(
        "u_UV",
        &rel.uv,
        ShaderParameterType::VEC4
      ));

      parameters.push_back(ShaderParameter(
        "u_Color",
        &rel.color,
        ShaderParameterType::COLOR
      ));
    },
    2
  ));
}